@@ -1,11 +1,12 @@
 from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
-from typing import Iterable, List, Mapping, Union
+from typing import Any, Iterable, List, Mapping, Optional, Union, overload
 
 import pywikibot
 
 from .constants import *
 from .dataclasses import *
+from .process_reason import *
 from .sentry import *
 from .transformers import *
 from .utils import *
@@ -202,6 +203,173 @@ class PropertyAdderBot(ABC):
             return False
         return True
 
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.missing_property, ProcessReason.missing_value],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: None = None,
+        context: None = None,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.different_rank],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: None = None,
+        context: DifferentRankContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.replace_value],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: None = None,
+        context: ReplaceValueContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.delete_values],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: None = None,
+        context: DeleteValuesContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[
+            ProcessReason.missing_qualifier_property,
+            ProcessReason.missing_qualifier_value,
+        ],
+        *,
+        claim: ExtraProperty,
+        qualifier: ExtraQualifier,
+        reference: None = None,
+        context: None = None,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.replace_qualifier_value],
+        *,
+        claim: ExtraProperty,
+        qualifier: ExtraQualifier,
+        reference: None = None,
+        context: ReplaceQualifierValueContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.delete_qualifier_values],
+        *,
+        claim: ExtraProperty,
+        qualifier: ExtraQualifier,
+        reference: None = None,
+        context: DeleteQualifierValuesContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.new_claim_from_qualifier],
+        *,
+        claim: ExtraProperty,
+        qualifier: ExtraQualifier,
+        reference: None = None,
+        context: NewClaimFromQualifierContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.missing_reference],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: ExtraReference,
+        context: None = None,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.merged_reference],
+        *,
+        claim: ExtraProperty,
+        qualifier: None = None,
+        reference: ExtraReference,
+        context: MergedReferenceContext,
+    ):
+        ...
+
+    @overload
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: Literal[ProcessReason.post_output],
+        *,
+        claim: None = None,
+        qualifier: None = None,
+        reference: None = None,
+        context: None = None,
+    ):
+        ...
+
+    def processed_hook(
+        self,
+        item: EntityPage,
+        reason: ProcessReason,
+        *,
+        claim: Optional[ExtraProperty] = None,
+        qualifier: Optional[ExtraQualifier] = None,
+        reference: Optional[ExtraReference] = None,
+        context: Optional[dict[str, Any]] = None,
+    ):
+        """Do processing whenever the item is modified. This method is called directly after the item is modified.
+
+        :param item: The item that was modified.
+        :param reason: The reason the item was modified.
+        :param claim: The main claim that was added or is having qualifiers/references added, defaults to None
+        :param qualifier: The qualifier that was modified, defaults to None
+        :param reference: The reference that was modified, defaults to None
+        :param context: Additional context with the operation, defaults to None
+        """
+        pass
+
     def process(self, output: Output, item: EntityPage) -> bool:
         """Processes the output from run_item.
 
@@ -222,7 +390,11 @@ class PropertyAdderBot(ABC):
                     if self.can_add_main_property(
                         extra_prop_data
                     ) and self.whitelisted_claim(extra_prop_data):
+                        # This is triggered if there are no statements for the property
                         add_claim_locally(item, new_claim)
+                        self.processed_hook(
+                            item, ProcessReason.missing_property, claim=extra_prop_data
+                        )
                         acted = True
                     else:
                         continue
@@ -230,12 +402,23 @@ class PropertyAdderBot(ABC):
                     for existing_claim in item.claims[property_id].copy():
                         existing_claim: pywikibot.Claim
                         if self.same_main_property(existing_claim, new_claim, item):
+                            # This is triggered if there is a statement for the property exactly matching the one we want to add
                             if new_claim.getRank() != existing_claim.getRank():
                                 if (
                                     self.whitelisted_claim(extra_prop_data)
                                     or self.config.copy_ranks_for_nonwhitelisted_main_properties
                                 ):
+                                    old_rank = existing_claim.getRank()
                                     existing_claim.rank = new_claim.getRank()
+                                    self.processed_hook(
+                                        item,
+                                        ProcessReason.different_rank,
+                                        claim=extra_prop_data,
+                                        context=DifferentRankContext(
+                                            existing_claim=existing_claim,
+                                            old_rank=old_rank,
+                                        ),
+                                    )
                                     acted = True
                             new_claim = extra_prop_data.claim = existing_claim
                             break
@@ -244,19 +427,55 @@ class PropertyAdderBot(ABC):
                                 extra_prop_data.replace_if_conflicting_exists
                                 and self.whitelisted_claim(extra_prop_data)
                             ):
+                                # This is triggered if `extra_prop_data.replace_if_conflicting_exists` is set to True
+                                # and this is the first statement with the property that is not exactly matching the one we want to add
+                                old_value = existing_claim.getTarget()
                                 existing_claim.setTarget(new_claim.getTarget())
                                 if new_claim.getRank() != existing_claim.getRank():
-                                    new_claim.rank = existing_claim.getRank()
+                                    existing_claim.rank = new_claim.getRank()
+                                    self.processed_hook(
+                                        item,
+                                        ProcessReason.different_rank,
+                                        claim=extra_prop_data,
+                                        context=DifferentRankContext(
+                                            existing_claim=existing_claim,
+                                            old_value=old_value,
+                                        ),
+                                    )
+                                original_new_claim = new_claim
                                 new_claim = extra_prop_data.claim = existing_claim
+                                self.processed_hook(
+                                    item,
+                                    ProcessReason.replace_value,
+                                    claim=extra_prop_data,
+                                    context=ReplaceValueContext(
+                                        existing_claim=existing_claim,
+                                        new_claim=original_new_claim,
+                                        old_value=old_value,
+                                    ),
+                                )
                                 if (
                                     len(item.claims[property_id]) > 1
                                     and extra_prop_data.delete_other_if_replacing
                                 ):
+                                    deleted = item.claims[property_id].copy()
+                                    deleted.remove(new_claim)
+                                    self.processed_hook(
+                                        item,
+                                        ProcessReason.delete_values,
+                                        claim=extra_prop_data,
+                                        context=DeleteValuesContext(
+                                            deleted_claims=deleted
+                                        ),
+                                    )
                                     item.claims[property_id] = [new_claim]
                                 acted = True
                                 break
                     else:
+                        # This code section triggers if there are statements for the property but none of them match the one we want to add
+                        # and we did not opt for replacement.
                         if extra_prop_data.skip_if_conflicting_language_exists and property_id in item.claims:  # type: ignore
+                            found_conflicting_language = False
                             for existing_claim in item.claims[property_id]:  # type: ignore
                                 existing_claim: pywikibot.Claim
                                 if isinstance(
@@ -264,24 +483,37 @@ class PropertyAdderBot(ABC):
                                     pywikibot.WbMonolingualText,
                                 ):
                                     lang_target: pywikibot.WbMonolingualText = existing_claim.getTarget()  # type: ignore
-                                    if lang_target.language == new_claim.getTarget().language:  # type: ignore
+                                    if lang_target.language == new_claim.getTarget().language and lang_target != new_claim.getTarget():  # type: ignore
+                                        found_conflicting_language = True
                                         break
                                 else:
+                                    # The existing claim is not a monolingual text, so we can't compare it to the new one
                                     continue
                             else:
+                                # If we're here, we did not find a conflicting language
                                 if self.can_add_main_property(
                                     extra_prop_data
                                 ) and self.whitelisted_claim(extra_prop_data):
+                                    self.processed_hook(
+                                        item,
+                                        ProcessReason.missing_value,
+                                        claim=extra_prop_data,
+                                    )
                                     add_claim_locally(item, new_claim)
                                     acted = True
                                 else:
                                     continue
+                            if found_conflicting_language:
+                                continue
                         elif extra_prop_data.skip_if_conflicting_exists:
                             continue
                         if self.can_add_main_property(
                             extra_prop_data
                         ) and self.whitelisted_claim(extra_prop_data):
                             add_claim_locally(item, new_claim)
+                            self.processed_hook(
+                                item, ProcessReason.missing_value, claim=extra_prop_data
+                            )
                             acted = True
                         else:
                             continue
@@ -294,6 +526,12 @@ class PropertyAdderBot(ABC):
                             self.whitelisted_claim(extra_prop_data)
                             or self.whitelisted_qualifier(extra_prop_data, qualifier)
                         ):
+                            self.processed_hook(
+                                item,
+                                ProcessReason.missing_qualifier_property,
+                                claim=extra_prop_data,
+                                qualifier=qualifier_data,
+                            )
                             add_qualifier_locally(new_claim, qualifier)
                             acted = True
                         else:
@@ -314,21 +552,49 @@ class PropertyAdderBot(ABC):
                                             )
                                         )
                                     ):
+                                        old_value = existing_qualifier.getTarget()
                                         existing_qualifier.setTarget(
                                             qualifier.getTarget()
                                         )
-                                        qualifier = existing_qualifier
+                                        self.processed_hook(
+                                            item,
+                                            ProcessReason.replace_qualifier_value,
+                                            claim=extra_prop_data,
+                                            qualifier=qualifier_data,
+                                            context=ReplaceQualifierValueContext(
+                                                existing_qualifier=existing_qualifier,
+                                                new_qualifier=qualifier,
+                                                old_value=old_value,
+                                            ),
+                                        )
+                                        qualifier = (
+                                            qualifier_data.claim
+                                        ) = existing_qualifier
                                         if (
                                             len(new_claim.qualifiers[qualifier_prop])
                                             > 1
                                             and qualifier_data.delete_other_if_replacing
                                         ):
+                                            deleted = new_claim.qualifiers[
+                                                qualifier_prop
+                                            ].copy()
+                                            deleted.remove(qualifier)
+                                            self.processed_hook(
+                                                item,
+                                                ProcessReason.delete_qualifier_values,
+                                                claim=extra_prop_data,
+                                                qualifier=qualifier_data,
+                                                context=DeleteQualifierValuesContext(
+                                                    deleted_claims=deleted
+                                                ),
+                                            )
                                             new_claim.qualifiers[qualifier_prop] = [
                                                 qualifier
                                             ]
                                         acted = True
                                         break
                             else:
+                                made_new_claim = False
                                 if qualifier_data.skip_if_conflicting_exists:
                                     continue
                                 elif (
@@ -336,14 +602,32 @@ class PropertyAdderBot(ABC):
                                     and self.whitelisted_claim(extra_prop_data)
                                 ):
                                     if self.can_add_main_property(extra_prop_data):
+                                        old_claim = new_claim
                                         new_claim = (
                                             extra_prop_data.claim
                                         ) = original_claim
                                         add_claim_locally(item, new_claim)
+                                        self.processed_hook(
+                                            item,
+                                            ProcessReason.new_claim_from_qualifier,
+                                            claim=extra_prop_data,
+                                            qualifier=qualifier_data,
+                                            context=NewClaimFromQualifierContext(
+                                                old_claim=old_claim
+                                            ),
+                                        )
                                         acted = True
+                                        made_new_claim = True
                                     else:
                                         continue
                                 add_qualifier_locally(new_claim, qualifier)
+                                if not made_new_claim:
+                                    self.processed_hook(
+                                        item,
+                                        ProcessReason.missing_qualifier_value,
+                                        claim=extra_prop_data,
+                                        qualifier=qualifier_data,
+                                    )
                                 acted = True
                 for extra_reference in extra_prop_data.extra_references:
                     compatible = False
@@ -361,12 +645,27 @@ class PropertyAdderBot(ABC):
                                 existing_reference,
                                 list(extra_reference.new_reference_props.values()),
                             ):
+                                self.processed_hook(
+                                    item,
+                                    ProcessReason.merged_reference,
+                                    claim=extra_prop_data,
+                                    reference=extra_reference,
+                                    context=MergedReferenceContext(
+                                        existing_reference=existing_reference
+                                    ),
+                                )
                                 acted = True
                             break
                     if not compatible and (
                         self.whitelisted_claim(extra_prop_data)
                         or self.whitelisted_reference(extra_prop_data, extra_reference)
                     ):
+                        self.processed_hook(
+                            item,
+                            ProcessReason.missing_reference,
+                            claim=extra_prop_data,
+                            reference=extra_reference,
+                        )
                         add_reference_locally(
                             new_claim, *extra_reference.new_reference_props.values()
                         )
@@ -374,8 +673,10 @@ class PropertyAdderBot(ABC):
         with start_span(
             op="post_output_process", description="Post Output Process Hook"
         ):
-            if self.post_output_process_hook(output, item) and not acted:
-                acted = True
+            if self.post_output_process_hook(output, item):
+                self.processed_hook(item, ProcessReason.post_output_process_hook)
+                if not acted:
+                    acted = True
         if acted:
             with start_span(op="pre_edit_process", description="Pre Edit Process Hook"):
                 self.pre_edit_process_hook(output, item)
